"""Transformers calculating string kernels."""
from __future__ import annotations

import collections
import logging
from collections import Counter
from typing import Callable, Dict, List

import numpy as np

logger = logging.getLogger(__name__)


def _multiset_kernel(
    x: np.ndarray,
    y: np.ndarray,
    callback: Callable,
) -> np.ndarray:
    """
    Helper function for all other kernels in this file.
    You probably don't want to use this method outside of this file.

    Args:
        x, y: a numpy array of documents
        callback: one of the kernel methods in this file
        **callback_kwargs: optional arguments to the callback

    Returns:
        a len(x) by len(y) matrix containing the kernel distances.
    """
    result = np.zeros((len(x), len(y)))
    x_counts: List[Dict] = [Counter(d) for d in x]
    y_counts: List[Dict] = [Counter(d) for d in y]
    for i, xc in enumerate(x_counts):
        for j, yc in enumerate(y_counts):
            result[i, j] = callback(xc, yc)
    return result


def presence_kernel(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Calculate the presence kernel, Ionescu & Popescu 2017.

    x and y are lists of documents. Each document must be an iterable of features that
    can be compared using ==.

    Returns:
        a matrix where each entry [i, j] represents the number of features that document
        x[i] and y[j] have in common.
    """
    return _multiset_kernel(x, y, lambda xc, yc: len(xc & yc))


def spectrum_kernel(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Calculate the spectrum kernel, Ionescu & Popescu 2017.

    x and y are lists of documents. Each document must be an iterable of features that
    can be compared using ==.

    Returns:
        a matrix where each entry [i, j] = sum(
           xc * yc for all xc, yc in (common features in x[i] and y[j])
        )
    """

    def inner(xc: collections.Counter, yc: collections.Counter) -> float:
        all_ngrams = set(xc.keys()).intersection(set(yc.keys()))
        return sum([xc[ngram] * yc[ngram] for ngram in all_ngrams])

    return _multiset_kernel(x, y, inner)


def intersection_kernel(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Calculate the intersection kernel, Ionescu & Popescu 2017.

    x and y are lists of documents. Each document must be an iterable of features that
    can be compared using ==.

    Returns:
        a matrix where each entry [i, j] = sum(
            min(xc, yc) for all xc, yc in (common features in x[i] and y[j])
        )
    """
    return _multiset_kernel(x, y, lambda xc, yc: sum((xc & yc).values()))


def pqgram_kernel(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Calculates a distance kernel based on the PQ-gram distance by Austen et al. 2010.
    """

    def pqgram_distance(xc: collections.Counter, yc: collections.Counter) -> float:
        union = sum((xc + yc).values())  # bag union
        intersection = sum((xc & yc).values())
        # eq. (2) and (3) of Augsten et al. 2010
        return (union - 2 * intersection) / (union - intersection)

    return _multiset_kernel(x, y, pqgram_distance)



import json
import os
import pickle
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from typing import Any, Dict, List

import langdetect
import pandas as pd
from sklearn.pipeline import make_pipeline
from tqdm import tqdm
from treegrams.transformers import TreeGramExtractor

import stanza
from tuhlbox import logger
from tuhlbox.stanza import StanzaToNltkTreesTransformer


class Contributor(ABC):
    def __init__(
        self,
        column_name: str,
        required_columns: List[str],
        flush_after_each_row: bool = False,
        overwrite: bool = False,
    ) -> None:
        self.column_name = column_name
        self.required_columns = required_columns
        self.flush_after_each_row = flush_after_each_row
        self.base_dir: str = ""
        self.overwrite = overwrite

    def contribute(self, csv_path: str) -> None:
        self.base_dir = os.path.dirname(csv_path)
        name = self.__class__.__name__
        df = pd.read_csv(csv_path)
        if not set(self.required_columns).issubset(set(df.columns)):
            raise Exception(f"not all column requirements for {name} are fulfilled.")

        if self.column_name in df.columns and not (
            self.overwrite or self.flush_after_each_row
        ):
            logger.info("skipping contributor %s, column exists", name)
            return
        logger.info(f"{name} is calculating column: " f"{self.column_name}")
        df = self.calculate(df)
        df.to_csv(csv_path)

    @abstractmethod
    def calculate(self, row: pd.DataFrame) -> pd.DataFrame:
        pass

    def read_subdir_file(self, path: str, mode: str = "text") -> Any:
        full_path = os.path.join(self.base_dir, path)
        if mode == "text":
            with open(full_path) as text_fh:
                return text_fh.read()
        if mode == "pickle":
            with open(full_path, "rb") as pickle_fh:
                return pickle.load(pickle_fh)

    def write_subdir_file(self, path: str, content: Any, mode: str = "text") -> None:
        full_path = os.path.join(self.base_dir, path)
        if mode == "text":
            with open(full_path, "w") as text_fh:
                text_fh.write(content)
        if mode == "pickle":
            with open(full_path, "wb") as pickle_fh:
                pickle.dump(content, pickle_fh)


class RowWiseContributor(Contributor):
    @abstractmethod
    def calculate_row(self, row: pd.Series) -> pd.Series:
        pass

    def calculate(self, df: pd.DataFrame) -> pd.DataFrame:
        new_rows = []
        for _, row in tqdm(df.iterrows(), total=df.shape[0]):
            new_rows.append(self.calculate_row(row))
        return pd.DataFrame(new_rows)


class LinguisticStatsContributor(RowWiseContributor):
    """
    Calculates the type/token ratio along with many other counts.

    Returns a dict with the distinct number of:
     - "upos": universal POS tags
     - "xpos": language-specific POS tags
     - "lemma": the word's lemma
     - "text": the word's text (type)
     - "deprel": the universal dependency relationship of this word to its parent
     - "tokens": total number of non-distinct tokens
     - "words": total number of non-distinct words
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            column_name="linguistic_stats", required_columns=["stanza"], **kwargs
        )

    def calculate_row(self, row: pd.Series) -> pd.Series:
        document = self.read_subdir_file(row["stanza"], "pickle")
        fields = ["upos", "xpos", "lemma", "text", "deprel"]
        sets = defaultdict(set)
        for word in document.iter_words():
            for field in fields:
                sets[field].add(getattr(word, field))
        result: Dict[str, Any] = {f: len(sets[f]) for f in fields}
        result["tokens"] = document.num_tokens
        result["words"] = document.num_words
        result["characters"] = len(f"{document.text}")
        result["sentence_lengths"] = [len(sent.tokens) for sent in document.sentences]
        row[self.column_name] = json.dumps(result)
        return row


class StanzaContributor(RowWiseContributor):
    def __init__(self, directory_name: str = "stanza", **kwargs: Any) -> None:
        super().__init__(
            column_name="stanza",
            required_columns=["text_raw", "language"],
            flush_after_each_row=True,
            **kwargs,
        )
        self.parsers: Dict[str, stanza.Pipeline] = {}
        self.directory_name = directory_name

    def calculate_row(self, row: pd.Series) -> pd.Series:
        lang = row["language"]
        input_filename = row["text_raw"]
        stanza_dir = os.path.join(self.base_dir, self.directory_name)
        if not os.path.isdir(stanza_dir):
            os.makedirs(stanza_dir)
        out_filename = os.path.splitext(os.path.basename(input_filename))[0] + ".pckl"
        out_filepath = os.path.join(self.directory_name, out_filename)
        full_out_filepath = os.path.join(self.base_dir, out_filepath)

        if not os.path.isfile(full_out_filepath) or self.overwrite:
            text = self.read_subdir_file(input_filename)
            if lang not in self.parsers:
                self.parsers[lang] = stanza.Pipeline(
                    lang=lang, use_gpu=False, processors="tokenize,pos,depparse,lemma"
                )
            document = self.parsers[lang](text)
            self.write_subdir_file(out_filepath, document, "pickle")
        row[self.column_name] = out_filepath
        return row


class LanguageDetectionContributor(RowWiseContributor):
    def __init__(self, min_lang_prob: float = 0.99) -> None:
        super().__init__(column_name="language", required_columns=["text_raw"])
        self.min_lang_prob = min_lang_prob

    def calculate_row(self, row: pd.Series) -> pd.Series:
        text = self.read_subdir_file(row["text_raw"])
        langdetect.detect_langs(text)
        language_scores = sorted(
            list(langdetect.detect_langs(text)), key=lambda x: x.prob, reverse=True
        )
        scores = [x.prob for x in language_scores]
        langs = [x.lang for x in language_scores]
        if scores[0] > self.min_lang_prob:
            row[self.column_name] = langs[0]
        else:
            row[self.column_name] = None
        return row


class CounterContributor(RowWiseContributor):
    def __init__(
        self,
        column_name: str = "dtgram_counter",
        directory_name: str = "dtgram_counters",
        **kwargs: Any,
    ) -> None:
        super().__init__(column_name=column_name, required_columns=["stanza"], **kwargs)
        self.column_name = column_name
        self.directory_name = directory_name
        self.dtgram_pipeline = make_pipeline(
            StanzaToNltkTreesTransformer(),
            TreeGramExtractor(),
        )

    def calculate_row(self, row: pd.Series) -> pd.Series:
        stanza_document = self.read_subdir_file(row["stanza"], mode="pickle")

        counter_dir = os.path.join(self.base_dir, self.directory_name)
        if not os.path.isdir(counter_dir):
            os.makedirs(counter_dir)
        out_filename = os.path.splitext(os.path.basename(row["stanza"]))[0] + ".pckl"
        out_filepath = os.path.join(self.directory_name, out_filename)
        full_out_filepath = os.path.join(self.base_dir, out_filepath)

        if not os.path.isfile(full_out_filepath) or self.overwrite:
            dt_grams = self.dtgram_pipeline.transform([stanza_document])[0]
            counter: Counter = Counter(dt_grams)
            self.write_subdir_file(out_filepath, counter, mode="pickle")

        row[self.column_name] = out_filepath
        return row


class ConstantContributor(Contributor):
    def __init__(self, column_name: str, value: Any) -> None:
        super().__init__(column_name=column_name, required_columns=[])
        self.column_name = column_name
        self.value = value

    def calculate(self, df: pd.DataFrame) -> pd.DataFrame:
        df[self.column_name] = [self.value] * df.shape[0]
        return df



"""Dataset manipulation scripts."""

import json
import logging
import os
import pickle
import shutil
import sys
import warnings
from copy import deepcopy
from glob import glob
from typing import List
from urllib.error import HTTPError
from urllib.request import urlretrieve

import click
import pandas as pd
import torch
from tqdm import tqdm  # type: ignore
from transformers import MarianMTModel, MarianTokenizer  # type: ignore

import stanza  # type: ignore

logger = logging.getLogger(__name__)

FEATURE_TEXT = "text_raw"
FEATURE_STANZA = "stanza"
FEATURE_CONSTITUENTS = "constituencies"
LANGUAGE_COLUMN = "language"
DATASET_CSV = "dataset.csv"
HF_URL = "https://s3.amazonaws.com/models.huggingface.co/bert/Helsinki-NLP"
FILES = [
    "config.json",
    "pytorch_model.bin",
    "source.spm",
    "target.spm",
    "tokenizer_config.json",
    "vocab.json",
]
MODEL_DIR = "translation_data"


@click.command(
    help="transforms corpora created with the reddit script to the dataframe format."
)
@click.argument("input_directory")
@click.argument("output_directory")
def reddit_to_common(input_directory: str, output_directory: str) -> None:
    """
    Transform corpora created with reddit scripts into dataframe format.

    This script is meant to be called via CLI. It will read a directory that
    contains a nested directory structure and produces a common dataset format
    using DataFrames. The result is a meta-information file called dataset.csv
    and a directory 'text_raw' which contains the raw text files.

    Args:
        input_directory: The directory that contains the topmost target
            directories (often, authors).

    Returns: Nothing.

    """
    records = []
    authors = sorted(os.listdir(input_directory))

    raw_directory = os.path.join(output_directory, "raw")
    if not os.path.isdir(raw_directory):
        os.makedirs(raw_directory)
    processed_directory = os.path.join(output_directory, "processed")
    if not os.path.isdir(processed_directory):
        os.makedirs(processed_directory)
    text_directory = os.path.join(processed_directory, FEATURE_TEXT)
    if not os.path.isdir(text_directory):
        os.makedirs(text_directory)

    meta_file = os.path.join(processed_directory, DATASET_CSV)

    for author in tqdm(authors):

        author_dir = os.path.join(input_directory, author)
        if not os.path.isdir(author_dir):
            logger.warning("not an author directory: %s", author_dir)
            continue
        languages = os.listdir(author_dir)
        for language in languages:
            language_dir = os.path.join(author_dir, language)
            if not os.path.isdir(language_dir):
                logger.warning("not a language directory: %s", language_dir)
                continue
            json_files = glob(language_dir + "/*.json")
            for json_file in json_files:
                name_ext = os.path.basename(json_file)
                name = os.path.splitext(name_ext)[0]
                text_name = f"{author}_{language}_{name}.txt"
                text_file = os.path.join(FEATURE_TEXT, text_name)
                full_text_file = os.path.join(processed_directory, text_file)
                with open(json_file) as i_f, open(full_text_file, "w") as o_f:
                    js = json.load(i_f)
                    o_f.write(js["body_clean"])
                    del js["body"]
                    del js["body_clean"]
                    js[FEATURE_TEXT] = text_file
                    js["group_field"] = language  # important for deepl_...
                    records.append(js)

        # move old author dir
    shutil.copytree(input_directory, raw_directory, dirs_exist_ok=True)

    df = pd.DataFrame.from_records(records)
    df.to_csv(meta_file, index=False)


@click.command(help="Reads dataset.csv + column name to produce stanza dir")
@click.argument("input-directory")
@click.option("-t", "--text-column-name", default=FEATURE_TEXT)
@click.option("-l", "--language-column-name", default=LANGUAGE_COLUMN)
@click.option("-o", "--overwrite", default=False)
@click.option("-out", "--output-column-name", default=FEATURE_STANZA)
def parse_dependency(
    input_directory: str,
    text_column_name: str,
    language_column_name: str,
    overwrite: bool,
    output_column_name: str,
) -> None:
    """
    Parse text files using the stanza parser.

    Args:
        input_directory: Directory containing the dataset.csv file, and a
            directory named <text_column_name>
        text_column_name: name of the column containing the raw text data.
            defaults to 'text_raw'.
        language_column_name:  name of the column where the language is saved.
            defaults to 'language'.
        overwrite: whether to re-parse existing files. If this is not set, this
            script will leave already parsed files untouched and won't parse
            the appropriate inputs.
        output_column_name: name of the output column in the meta-data file.
            defaults to 'stanza'.

    Returns: Nothing, this is a cli script.

    """
    warnings.simplefilter("ignore")  # stanza is very 'loud'.
    main_dataset_file = os.path.join(input_directory, DATASET_CSV)
    df = pd.read_csv(main_dataset_file)

    # check output directory
    out_dir = os.path.join(input_directory, output_column_name)
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    df[output_column_name] = [
        os.path.join(
            output_column_name, os.path.splitext(os.path.basename(f))[0] + ".pckl"
        )
        for f in df[text_column_name]
    ]

    def should_write(f: str) -> bool:
        if overwrite:
            return True
        if os.path.isfile(f):
            return os.path.getsize(f) == 0
        else:
            return True

    tuples = [
        (in_file, out_file, language)
        for (in_file, out_file, language) in zip(
            df[text_column_name], df[output_column_name], df[language_column_name]
        )
        if should_write(os.path.join(input_directory, out_file))
    ]
    if not tuples:
        logger.warning("no files remaining, skipping calculation")
        return

    parsers = {}
    for in_file, out_file, language in tqdm(tuples):
        if language.endswith("_to_en"):
            language = "en"
        if language not in parsers:
            parsers[language] = stanza.Pipeline(lang=language, use_gpu=False)
        parser = parsers[language]

        full_path = os.path.join(input_directory, in_file)
        with open(full_path) as in_fh:
            content = in_fh.read()
            try:
                parsed = parser(content)
            except RuntimeError as e:
                logger.error("error on parsing", full_path)
                logger.error(content)
                raise e
            with open(os.path.join(input_directory, out_file), "wb") as out_fh:
                pickle.dump(parsed, out_fh)
    logger.info("writing %s", main_dataset_file)
    df.to_csv(main_dataset_file, index=False)


@click.command(help="reads dataset.csv, produces constituencies directory")
@click.argument("input-directory")
@click.option("-t", "--text-column-name", default=FEATURE_TEXT)
@click.option("-l", "--language-column-name", default=LANGUAGE_COLUMN)
def parse_constituency(
    input_directory: str, text_column_name: str, language_column_name: str
) -> None:
    """
    Ignore this method for the time being.

    This is no longer working unless the common file transformers are replaced.
    """
    print(input_directory, text_column_name, language_column_name)
    pass
    # meta = os.path.join(input_directory, DATASET_CSV)
    #
    # df = pd.read_csv(meta)
    #
    # out_dir = os.path.join(input_directory, FEATURE_CONSTITUENTS)
    # if not os.path.isdir(out_dir):
    #     os.makedirs(out_dir)
    #
    # sub_dfs = {}
    #
    # columns = [text_column_name, language_column_name]
    # for language, sub_df in df[columns].groupby(language_column_name):
    #     logger.info('parsing %s', language)
    #
    #     infiles = list(sub_df[text_column_name].values)
    #     # the path to the text file points to a subfolder
    #     outfiles = PathTransformer('../' + FEATURE_CONSTITUENTS,
    #                                '.json').transform(infiles)
    #     pairs = [(i_f, o_f) for (i_f, o_f) in zip(infiles, outfiles)
    #              if not os.path.isfile(o_f)]
    #     if not pairs:
    #         logger.warning('no files remaining, skipping calculation')
    #         return
    #
    #     input_filenames, output_filenames = list(zip(*pairs))
    #
    #
    #
    #     make_pipeline(
    #         FileReader(),
    #         CoreNLPTreeTransformer(
    #             language=language,
    #             port=CORENLP_PORT,
    #             model_dir=CORENLP_MODELS,
    #             properties_dir=CORENLP_PROPERTIES),
    #         FileWriter(output_filenames, writemode='json'),
    #     ).transform(input_filenames)
    #
    #     sub_df[FEATURE_CONSTITUENTS] = output_filenames
    #     sub_dfs[language] = sub_df
    #
    # pd.concat(sub_dfs, axis=0).to_csv(DATASET_CSV + '2', index=False)


class Translator:
    """Simple translator wrapper for Hugging Face models."""

    def __init__(
        self,
        source_language: str,
        target_language: str,
        model_dir: str = "translation_models",
        force_download: bool = False,
    ):
        """
        Initialize the translator.

        Args:
            source_language: two-letter code for the source language
            target_language: two-letter code for the target language
            model_dir: directory where to store the cached translation models
            force_download: if true, overwrite existing cached models
        """
        self.source_language = source_language
        self.target_language = target_language
        self.model_dir = model_dir
        self.force_download = force_download

        self.model_name = f"opus-mt-{self.source_language}-" f"{self.target_language}"
        self.model_dir = os.path.join(self.model_dir, self.model_name)

        if not os.path.isdir(self.model_dir) or self.force_download:
            self._download_language_model()

        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        logger.info("running on device: %s", self.device)

        self.model = MarianMTModel.from_pretrained(self.model_dir).to(
            torch.device(self.device)
        )
        self.tokenizer = MarianTokenizer.from_pretrained(self.model_dir)

    def _download_language_model(self) -> None:
        if os.path.isdir(self.model_dir):
            shutil.rmtree(self.model_dir)
        os.makedirs(self.model_dir)
        for f in FILES:
            file_url = os.path.join(HF_URL, self.model_name, f)
            file_path = os.path.join(self.model_dir, f)
            try:
                logger.info("downloading %s", file_url)
                urlretrieve(file_url, file_path)
            except HTTPError as e:
                logger.error(
                    "Error retrieving model from url."
                    "Please confirm model exists: %s",
                    e,
                )
                sys.exit(1)

    def translate(self, sentences: List[str]) -> List[str]:
        """
        Translate sentences for source to target language.

        Args:
            sentences: list of strings to translate. Each sentence will be
                translated separately for memory reasons.

        Returns: list of translated sentences.

        """
        result = []
        for sentence in tqdm(sentences, leave=False):
            batch = self.tokenizer.prepare_seq2seq_batch(
                src_texts=[sentence], return_tensors="pt"
            )
            translated = self.model.generate(**batch.to(self.device))
            result += self.tokenizer.batch_decode(translated, skip_special_tokens=True)
        return result


@click.command(help="translates data")
@click.argument("input-dir")
@click.argument("source")
@click.argument("target")
def translate(
    input_dir: str,
    source: str,
    target: str,
) -> None:
    """
    Translate common datafrarme corpora.

    The corpora that is to be translated should already contain the 'stanza'
    column of parsed sentences, as this will have better sentence segmentation.

    Args:
        input_dir: directory containing the dataset.csv file.
        source: two-letter source language code
        target: two-letter target language code

    Returns: Nothing, this is a CLI script.

    """
    key = f"marianmt_{source}_to_{target}"
    df = pd.read_csv(os.path.join(input_dir, "dataset.csv"))
    df = df.drop(columns=[c for c in df.columns if c.startswith("Unnamed: 0")])
    sub_df = df[df["language"] == source]

    logger.info("loading translation model")
    translator = Translator(source, target)

    new_rows = []

    for index, row in tqdm(sub_df.iterrows(), total=sub_df.shape[0]):
        new_row = deepcopy(row)
        old_path = os.path.join(input_dir, row["stanza"])
        old_name = os.path.splitext(os.path.basename(old_path))[0]
        new_name = old_name + "__" + key
        new_path_part = os.path.join("text_raw", new_name + ".txt")
        new_path = os.path.join(input_dir, new_path_part)
        new_row["text_raw"] = new_path_part
        new_row["stanza"] = None
        new_row["language"] = key

        new_dir = os.path.dirname(new_path)
        if not os.path.isdir(new_dir):
            os.makedirs(new_dir)

        # sometimes, empty files are created when interrupting a process
        if os.path.isfile(new_path) and os.stat(new_path).st_size == 0:
            os.unlink(new_path)

        # don't re-compute stuff we already have
        if os.path.isfile(new_path):
            continue

        with open(old_path, "rb") as i_f, open(new_path, "w") as o_f:
            document = pickle.load(i_f)
            sentences = [sentence.text for sentence in document.sentences]
            translations = translator.translate(sentences)
            o_f.write("\n".join(translations))

        new_rows.append(new_row)

    df2 = pd.DataFrame.from_records(new_rows)
    df3 = pd.concat([df, df2])
    df3.to_csv(os.path.join(input_dir, "dataset.csv"), index=False)



"""Basic RNN model."""
from typing import Any

import torch
from torch import nn


class RNNClassifier(nn.Module):
    def __init__(
        self,
        n_classes: int,
        embedding_dim: int = 128,
        rec_layer_type: str = "lstm",
        num_units: int = 128,
        num_layers: int = 2,
        dropout: float = 0.0,
        vocab_size: int = 1000,
    ):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.rec_layer_type = rec_layer_type.lower()
        self.num_units = num_units
        self.num_layers = num_layers
        self.dropout = dropout
        self.n_classes = n_classes

        self.emb = nn.Embedding(vocab_size + 1, embedding_dim=self.embedding_dim)

        rec_layer = {"lstm": nn.LSTM, "gru": nn.GRU}[self.rec_layer_type]
        # We have to make sure that the recurrent layer is batch_first,
        # since sklearn assumes the batch dimension to be the first
        self.rec = rec_layer(
            self.embedding_dim,
            self.num_units,
            num_layers=num_layers,
            batch_first=True,
        )

        self.output = nn.Linear(self.num_units, self.n_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        embeddings = self.emb(x)
        # from the recurrent layer, only take the activities from the
        # last sequence step
        if self.rec_layer_type == "gru":
            _, rec_out = self.rec(embeddings)
        else:
            _, (rec_out, _) = self.rec(embeddings)
        rec_out = rec_out[-1]  # take output of last RNN layer
        drop_out = nn.Dropout(p=self.dropout)(rec_out)
        # Remember that the final non-linearity should be softmax, so
        # that our predict_proba method outputs actual probabilities!
        linear_out = self.output(drop_out)
        out = nn.Softmax(dim=-1)(linear_out)
        return out



"""Generic Wrapper for Transformer-Based models."""
from __future__ import annotations

from typing import Any, List, Union

import numpy as np
import pandas as pd
from simpletransformers.classification import ClassificationModel
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder


class SimpletransformersBertModel(BaseEstimator, ClassifierMixin):
    """Generic Wrapper for Transformer Models."""

    def __init__(
        self,
        model_type: str = "distilbert",
        pretrained_model: str = "distilbert-base-uncased",
        output_dir: str = "outputs/",
        cache_dir: str = "cache/",
        fp16: bool = False,
        fp16_opt_level: str = "O1",
        max_seq_length: int = 256,
        train_batch_size: int = 8,
        eval_batch_size: int = 8,
        gradient_accumulation_steps: int = 1,
        num_train_epochs: int = 1,
        weight_decay: int = 0,
        learning_rate: float = 4e-5,
        adam_epsilon: float = 1e-8,
        warmup_ratio: float = 0.06,
        warmup_steps: int = 0,
        max_grad_norm: float = 1.0,
        logging_steps: int = 0,
        evaluate_during_training: bool = False,
        save_steps: int = -1,
        save_eval_checkpoints: bool = False,
        save_model_every_epoch: bool = False,
        no_cache: bool = False,
        no_save: bool = False,
        eval_all_checkpoints: bool = True,
        use_tensorboard: bool = False,
        overwrite_output_dir: bool = False,
        reprocess_input_data: bool = True,
        use_cuda: bool = False,
        n_gpu: int = 1,
        silent: bool = False,
        use_multiprocessing: bool = True,
        sliding_window: bool = True,
    ):
        """Look at the documentation of simpletransformers for details."""
        self.model_type = model_type
        self.pretrained_model = pretrained_model
        self.output_dir = output_dir
        self.cache_dir = cache_dir
        self.fp16 = fp16
        self.fp16_opt_level = fp16_opt_level
        self.max_seq_length = max_seq_length
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size
        self.gradient_accumulation_steps = gradient_accumulation_steps
        self.num_train_epochs = num_train_epochs
        self.weight_decay = weight_decay
        self.learning_rate = learning_rate
        self.adam_epsilon = adam_epsilon
        self.warmup_ratio = warmup_ratio
        self.warmup_steps = warmup_steps
        self.max_grad_norm = max_grad_norm
        self.logging_steps = logging_steps
        self.evaluate_during_training = evaluate_during_training
        self.save_steps = save_steps
        self.eval_all_checkpoints = eval_all_checkpoints
        self.use_tensorboard = use_tensorboard
        self.overwrite_output_dir = overwrite_output_dir
        self.reprocess_input_data = reprocess_input_data
        self.use_cuda = use_cuda
        self.silent = silent
        self.use_multiprocessing = use_multiprocessing
        self.sliding_window = sliding_window
        self.save_eval_checkpoints = save_eval_checkpoints
        self.save_model_every_epoch = save_model_every_epoch
        self.no_cache = no_cache
        self.no_save = no_save

    def fit(
        self,
        x: Union[List[str], np.ndarray],
        y: Union[List[str], np.ndarray],
        *_args: Any,
        **_kwargs: Any,
    ) -> SimpletransformersBertModel:
        """Fit the model."""
        self.label_encoder = LabelEncoder()
        y = self.label_encoder.fit_transform(y)
        targets = set(y)
        n_classes = len(targets)
        self.classes_ = targets

        # sklearn requires x to always be two-dimensional.
        # however, older plans might use one-dimensional xs, and the remainder of this
        # script will use only one dimension.
        if type(x[0]) is not str:
            x = x[:, 0]  # type: ignore

        for target in targets:
            if target >= n_classes:
                raise Exception(
                    f"target {target} is larger than n_classes ({n_classes}). "
                    "this is not supported by pytorch and will cause a very "
                    "hard to debug error."
                )

        model_args = {
            "model_type": self.model_type,
            "pretrained_model": self.pretrained_model,
            "output_dir": self.output_dir,
            "cache_dir": self.cache_dir,
            "fp16": self.fp16,
            "fp16_opt_level": self.fp16_opt_level,
            "max_seq_length": self.max_seq_length,
            "train_batch_size": self.train_batch_size,
            "eval_batch_size": self.eval_batch_size,
            "gradient_accumulation_steps": self.gradient_accumulation_steps,
            "num_train_epochs": self.num_train_epochs,
            "weight_decay": self.weight_decay,
            "learning_rate": self.learning_rate,
            "adam_epsilon": self.adam_epsilon,
            "warmup_ratio": self.warmup_ratio,
            "warmup_steps": self.warmup_steps,
            "max_grad_norm": self.max_grad_norm,
            "logging_steps": self.logging_steps,
            "evaluate_during_training": self.evaluate_during_training,
            "save_steps": self.save_steps,
            "eval_all_checkpoints": self.eval_all_checkpoints,
            "use_tensorboard": self.use_tensorboard,
            "overwrite_output_dir": self.overwrite_output_dir,
            "reprocess_input_data": self.reprocess_input_data,
            "silent": self.silent,
            "use_multiprocessing": self.use_multiprocessing,
            "dataloader_num_workers": 0,
            "sliding_window": self.sliding_window,
            "save_eval_checkpoints": self.save_eval_checkpoints,
            "save_model_every_epoch": self.save_model_every_epoch,
            "no_save": self.no_save,
            "no_cache": self.no_cache,
        }

        self.model = ClassificationModel(
            self.model_type,
            self.pretrained_model,
            num_labels=n_classes,
            args=model_args,
            use_cuda=self.use_cuda,
        )
        df = pd.DataFrame(dict(text=x, labels=y))
        self.model.train_model(df)
        return self

    def predict(
        self,
        x: Union[List[str], np.ndarray],
        *_args: Any,
        **_kwargs: Any,
    ) -> Union[List[str], np.ndarray]:
        """Predict unseen documents."""
        # sklearn requires x to always be two-dimensional.
        # however, older plans might use one-dimensional xs.
        if type(x[0]) is not str:
            x = x[:, 0]  # type: ignore
        result, raw = self.model.predict(x)
        result = self.label_encoder.inverse_transform(result)
        return result



"""Transformers that calculate Doc2Vec embeddings using GenSim."""
from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import Any, List, Union

import numpy as np
from gensim.models.doc2vec import Doc2Vec, TaggedDocument  # type: ignore
from sklearn.base import BaseEstimator, TransformerMixin  # type: ignore

logging.getLogger("gensim").setLevel(logging.WARNING)
logging.getLogger("gensim").propagate = False


def _sanity_check(documents: List[List[str]]) -> None:
    """
    Check data for basic properties.

    Checks if the data is not-empty, and each document is a list of strings
    representing sentences.
    """
    if len(documents) < 1:
        raise ValueError("empty document set provided")

    if type(documents[0]) == str or not isinstance(documents[0], Iterable):
        raise TypeError("this transformer only works on pre-split data.")


class Doc2VecTransformer(TransformerMixin, BaseEstimator):
    """
    Transforms documents/sentences into a fixed-dimensional vector space.

    input: a list of list of strings
        each document is represented as a list of strings - those can be
        sentences, or tweets, or ...
    output: a list of list of numbers
        each document ist returned as the embedding vector.
    """

    def __init__(
        self,
        learning_rate: float = 0.02,
        epochs: int = 20,
        vector_size: int = 100,
        alpha: float = 0.025,
        min_alpha: float = 0.00025,
        min_count: int = 2,
        distributed_memory: int = 1,
        workers: int = 1,
    ):
        """
        Initialize Doc2Vec Transformer.

        Args:
            learning_rate: Learning rate of gensim model
            epochs: number of epochs to train the gensim model
            vector_size: final vector size
            alpha: alpha of the gensim model
            min_alpha: min_alpha of the gensim model
            min_count: min number of occurrences for each word
            distributed_memory: whether to use distributed memory model or not
            workers: number of threads
        """
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.vector_size = vector_size
        self.alpha = alpha
        self.min_alpha = min_alpha
        self.min_count = min_count
        self.distributed_memory = distributed_memory
        self.workers = workers

        self.model = Doc2Vec(
            vector_size=vector_size,
            alpha=alpha,
            min_alpha=min_alpha,
            min_count=min_count,
            dm=distributed_memory,
            workers=workers,
        )

    def fit(
        self,
        documents: List[List[str]],
        labels: Union[List[str], np.ndarray] = None,
        **fit_params: Any
    ) -> Doc2VecTransformer:
        """Fit the model by learning the training corpus."""
        documents = [[str(x) for x in document] for document in documents]

        _sanity_check(documents)
        tagged_x = [
            TaggedDocument(words=row, tags=[]) for _, row in enumerate(documents)
        ]
        self.model.build_vocab(tagged_x)
        self.model.train(
            tagged_x, total_examples=self.model.corpus_count, epochs=self.epochs
        )
        return self

    def transform(self, documents: List[List[str]]) -> np.ndarray:
        """Infer the vectors for documents."""
        documents = [[str(x) for x in document] for document in documents]
        _sanity_check(documents)
        vectors = [self.model.infer_vector(doc) for doc in documents]
        return np.array(vectors)



"""Transformer computing second-order attributes."""
from __future__ import annotations

import logging
import math
from collections import defaultdict
from functools import partial
from typing import Any, DefaultDict, Dict, Iterable, List

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

logger = logging.getLogger(__name__)


class SubFrequencyVectorizer(BaseEstimator, TransformerMixin):
    """Transformer computing second-order attributes."""

    def __init__(self) -> None:
        """Initialize the model."""
        self.t: DefaultDict[str, np.ndarray] = defaultdict(
            partial(np.ndarray, 0)
        )  # type:ignore

    def fit(
        self, X: List[str], y: Iterable[str], **_fit_params: Any
    ) -> SubFrequencyVectorizer:
        """Fit data to model."""
        words = set()

        logger.info("starting phase 0")
        documents = defaultdict(list)
        for document, target in zip(X, y):
            documents[target].append(document)
            for word in document.split():
                words.add(word)

        targets = set(y)
        tp: Dict[str, Dict[str, float]] = defaultdict(lambda: defaultdict(float))
        weights: Dict[str, Dict[str, float]] = defaultdict(lambda: defaultdict(float))

        # equation 1: weight calculation
        for target, target_docs in documents.items():
            for document in target_docs:
                doc_words = document.split()
                for doc_word in doc_words:
                    tf = doc_words.count(doc_word)
                    length = len(doc_words)
                    weights[target][doc_word] += math.log2(1 + tf / length)

        # equation 2.2: normalization
        for target in targets:
            for word in words:
                norm_term = 0.0
                for k in targets:
                    norm_term += weights[k][word]
                tp[target][word] = weights[target][word] / norm_term

        for word in words:
            self.t[word] = np.array([tp[target][word] for target in targets])

        return self

    def transform(self, X: Iterable[str], _y: Any = None) -> np.ndarray:
        """Transform data due to previously learned frequencies."""
        result: List[np.ndarray] = []
        for k in X:
            document_sum: np.ndarray = np.array([])
            doc_words = k.split()
            for j in doc_words:
                if j not in self.t:
                    continue
                if document_sum.shape[0] == 0:
                    document_sum = np.zeros(self.t[j].shape)
                tf = doc_words.count(j)
                document_sum += self.t[j] * tf / len(doc_words)
            result.append(document_sum)
        return np.array(result)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tuhlbox']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'dstoolbox>=0.10.1,<0.11.0',
 'gensim>=4.0.1,<5.0.0',
 'langdetect>=1.0.9,<2.0.0',
 'simpletransformers>=0.51.1,<0.52.0',
 'sklearn>=0.0,<0.1',
 'skorch>=0.10.0,<0.11.0',
 'stanza>=1.1.1,<2.0.0',
 'torch>=1.7.0,<2.0.0',
 'treegrams>=0.1.0,<0.2.0']

entry_points = \
{'console_scripts': ['contribute = tuhlbox.cli:run_contributor',
                     'parse_dependency = tuhlbox.cli:parse_dependency',
                     'reddit_to_common = tuhlbox.cli:reddit_to_common',
                     'translate = tuhlbox.cli:translate']}

setup_kwargs = {
    'name': 'tuhlbox',
    'version': '0.4.14',
    'description': 'Personal toolbox of language processing models.',
    'long_description': None,
    'author': 'Benjamin Murauer',
    'author_email': 'b.murauer@posteo.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git.uibk.ac.at/csak8736/tuhlbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)
